#include "vfsCopyMove.h"
#include "progress.h"
#include <PalmOS.h>
#include <VfsMgr.h>

#define MEM_CHUNK_SIZE				512		//slow but low on memory use

static char* vfsCopyMovePrvPathConcat(const char *first, const char *second)
{
	UInt16 firstLen = StrLen(first);
	char *dst;
	
	dst = MemPtrNew(firstLen + StrLen(second) + 2);
	if (!dst)
		return NULL;
	
	StrCopy(dst, first);
	if (!firstLen || first[firstLen - 1] != '/') {
	
		dst[firstLen] = '/';
		firstLen++;
	}
	
	StrCopy(dst + firstLen, second);
	
	return dst;
}

static Err vfsCopyMovePrvCopyData(FileRef fFrom, FileRef fTo, char *buf, struct Progress *progress)
{
	UInt32 size, done, now, done2, now2;
	Err e;
	
	e = VFSFileSize(fFrom, &size);
	if (e != errNone)
		return e;
	
	for (done = 0; done < size; done += now) {
		
		progressUpdate(progress, done, size);
		
		now = size - done;
		if (now > MEM_CHUNK_SIZE)
			now = MEM_CHUNK_SIZE;
		
		e = VFSFileRead(fFrom, now, buf, &now);
		if (e != errNone)
			return e;
		
		for (done2 = 0; done2 < now; done2 += now2) {
			
			e = VFSFileWrite(fTo, now - done2, buf + done2, &now2);
			if (e != errNone)
				return e;
		}
	}
	
	return errNone;
}

static Err vfsCopyMovePrvDo(UInt16 fromVrn, const char *fromFullPath, UInt16 toVrn, const char *toPath, Boolean isMove, FileInfoType* fit)
{
	FileRef fFrom, fTo;
	const char *name;
	char *dst, *src;
	UInt32 attr;
	Err e;
	
	//find name
	for (name = fromFullPath + StrLen(fromFullPath); name > fromFullPath && name[-1] != '/'; name--);
	
	//get info
	e = VFSFileOpen(fromVrn, fromFullPath, vfsModeRead, &fFrom);
	if (e != errNone)
		goto out;
	
	e = VFSFileGetAttributes(fFrom, &attr);
	if (e != errNone)
		goto out_close_from;

	if (attr & vfsFileAttrVolumeLabel) {
		e = sysErrParamErr;
		goto out_close_from;
	}

	//calc dst path
	dst = vfsCopyMovePrvPathConcat(toPath, name);
	if (!dst) {
		e = memErrNotEnoughSpace;
		goto out_close_from;
	}

	if (attr & vfsFileAttrDirectory) {		//recursively copy dir
		
		UInt32 dei = vfsIteratorStart;
		
		//create dst dir
		e = VFSDirCreate(toVrn, dst);
		if (e != errNone) {
			
			MemPtrFree(dst);
			goto out_close_from;
		}
		
		//open to set attrs
		e = VFSFileOpen(toVrn, dst, vfsModeWrite, &fTo);
		MemPtrFree(dst);
		if (e != errNone)
			goto out_close_from;
		
		//set attrs
		e = VFSFileSetAttributes(fTo, attr &~ (vfsFileAttrDirectory | vfsFileAttrLink));
		if (e != errNone)
			goto out_close_from;
		
		(void)VFSFileClose(fTo);
		
		//copy each file (we cannot delete as that would ruin enumeration)
		for (dei = vfsIteratorStart; dei != vfsIteratorStop &&  (e = VFSDirEntryEnumerate (fFrom, &dei, fit)) == errNone;) {
			
			src = vfsCopyMovePrvPathConcat(fromFullPath, fit->nameP);
			if (!src) {
				e = memErrNotEnoughSpace;
				goto out_close_from; 
			}
			
			dst = vfsCopyMovePrvPathConcat(toPath, name);
			if (!dst) {
				e = memErrNotEnoughSpace;
				MemPtrFree(src);
				goto out_close_from; 
			}
			
			e = vfsCopyMovePrvDo(fromVrn, src, toVrn, dst, false /* treat as copy since we'll delete the things */, fit);
			MemPtrFree(src);
			MemPtrFree(dst);
			
			if (e != errNone)
				goto out_close_from;
		}
		
		if (e == expErrEnumerationEmpty)
			e = errNone;
		else if (e != errNone)
			goto out_close_from;	
		
		//now delete files if this was a move
		do {
			dei = vfsIteratorStart;
			
			if (expErrEnumerationEmpty == VFSDirEntryEnumerate (fFrom, &dei, fit))
				break;
			
			src = vfsCopyMovePrvPathConcat(fromFullPath, fit->nameP);
			if (!src) {
				e = memErrNotEnoughSpace;
				goto out_close_from; 
			}
			
			e = VFSFileDelete(fromVrn, src);
			MemPtrFree(src);
			
		} while (e == errNone);
		
	}
	else {
		struct Progress *progress;

		//create dst file
		e = VFSFileOpen(toVrn, dst, vfsModeWrite | vfsModeCreate, &fTo);
		MemPtrFree(dst);
		if (e != errNone)
			goto out_close_from;
		
		//set attrs
		e = VFSFileSetAttributes(fTo, attr &~ (vfsFileAttrDirectory | vfsFileAttrLink));
		if (e != errNone)
			goto out_close_from;
		
		//do the copy
		progress = progressStart(fromFullPath);
		e = vfsCopyMovePrvCopyData(fFrom, fTo, fit->nameP, progress);
		progressFinish(progress);
		VFSFileClose(fTo);
	}


out_close_from:
	VFSFileClose(fFrom);

out:
	
	//delete src if was move and no error (done here after close)
	if (e == errNone && isMove)
		e = VFSFileDelete(fromVrn, fromFullPath);

	return e;
}


Err vfsCopyMove(UInt16 fromVrn, const char *fromFullPath, UInt16 toVrn, const char *toPath, Boolean isMove)
{
	FileInfoType fi;
	Err e;
	
	fi.nameP = MemPtrNew((fi.nameBufLen = MEM_CHUNK_SIZE - 1) + 1);
	if (!fi.nameP)
		return memErrNotEnoughSpace;
	
	e = vfsCopyMovePrvDo(fromVrn, fromFullPath, toVrn, toPath, isMove, &fi);
	
	MemPtrFree(fi.nameP);
	
	return e;
}