#define NON_INTERNATIONAL		//to avoid deprecation warning
#include <CharAttr.h>
#include <CharLatin.h>
#undef NON_INTERNATIONAL

#include <PalmOS.h>


struct BufferState {
	UInt8 *buf;
	UInt16 len, pos;
};

#define RW_ERR_INVAL_CHAR				-1
#define RW_ERR_NOT_ENOUGH_BYTES			-2

static const Int16 mPalmLatinHighChars[] = {		//others are identity
	0x20ac, RW_ERR_INVAL_CHAR, 0x201a, 0x0192, 0x201e, 0x2026, 0x2020, 0x2021, 0x02c6, 0x2030, 0x0160, 0x2039, 0x0152, RW_ERR_INVAL_CHAR, 0x0142, RW_ERR_INVAL_CHAR, 
	RW_ERR_INVAL_CHAR, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014, 0x02dc, 0x2122, 0x0161, 0x203a, 0x0153, RW_ERR_INVAL_CHAR, 0x017e, 0x0178,
};



static Int32 myTxtPrvReadAscii(struct BufferState *buf)	//read char.
{
	UInt8 chr;
	
	if (buf->pos >= buf->len)
		return RW_ERR_NOT_ENOUGH_BYTES;
	
	chr = buf->buf[buf->pos++];
	
	return (!chr || (chr >= 0x20 && chr < 0x80)) ? chr : RW_ERR_INVAL_CHAR;
}

static Int32 myTxtPrvReadPalmLatin(struct BufferState *buf)	//read char.
{
	UInt8 chr;
	
	if (buf->pos >= buf->len)
		return RW_ERR_NOT_ENOUGH_BYTES;
	
	chr = buf->buf[buf->pos++];
	
	if (!chr)
		return 0;
	else if (chr < 0x20)
		return RW_ERR_INVAL_CHAR;
	else if (chr < 0x80)
		return chr;
	else if (chr < 0xa0)
		return mPalmLatinHighChars[chr - 0x80];
	else
		return chr;
}

static Int32 myTxtPrvReadUCS2(struct BufferState *buf)
{
	UInt16 ret;
	
	if (buf->len - buf->pos < 2)
		return RW_ERR_NOT_ENOUGH_BYTES;
	
	ret = buf->buf[buf->pos++];
	ret <<= 8;
	ret += buf->buf[buf->pos++];
	
	return (UInt32)ret;
}

static Int32 myTxtPrvReadUTF8(struct BufferState *buf)
{
	UInt8 v;
	
	if (buf->pos >= buf->len)
		return RW_ERR_NOT_ENOUGH_BYTES;
	
	v = buf->buf[buf->pos];		//get len right away and make sure we have the space
	if (!(v & 0x80)) {			//ascii
		
		buf->pos++;
		return v;
	}
	if (!(v & 0x40)) {			//unexpected continuation
		
		buf->pos++;				//consume it
		return RW_ERR_INVAL_CHAR;
	}
	if (!(v & 0x20)) {			//2-byte char
		
		UInt16 t = v & 0x1f;
		
		if (buf->len - buf->pos < 2)
			return RW_ERR_NOT_ENOUGH_BYTES;
		
		t = (t << 6) | (buf->buf[buf->pos++] & 0x3f);
		
		if (t < 0x80)	//over-long encoding
			return RW_ERR_INVAL_CHAR;
		
		return t;
	}
	if (!(v & 0x10)) {			//3-byte char
		
		UInt16 t = v & 0x0f;
		
		if (buf->len - buf->pos < 3)
			return RW_ERR_NOT_ENOUGH_BYTES;
		
		t = (t << 6) | (buf->buf[buf->pos++] & 0x3f);
		t = (t << 6) | (buf->buf[buf->pos++] & 0x3f);
		
		if (t < 0x800)	//over-long encoding
			return RW_ERR_INVAL_CHAR;
		
		return t;
	}
	if (!(v & 0x08)) {			//4-byte char
		
		UInt32 t = v & 0x07;
		
		if (buf->len - buf->pos < 3)
			return RW_ERR_NOT_ENOUGH_BYTES;
		
		t = (t << 6) | (buf->buf[buf->pos++] & 0x3f);
		t = (t << 6) | (buf->buf[buf->pos++] & 0x3f);
		t = (t << 6) | (buf->buf[buf->pos++] & 0x3f);
		
		if (t < 0x10000 || t >= 0x110000)	//over-long or too-high-value encoding
			return RW_ERR_INVAL_CHAR;
		
		return t;
	}
	
	//invalid char
	buf->pos++;				//consume it
	return RW_ERR_INVAL_CHAR;
}

static Int32 myTxtPrvReadUTF16_generic(struct BufferState *buf, Boolean isLE)
{
	UInt16 ret, part2;
	
	if (buf->len - buf->pos < 2)
		return RW_ERR_NOT_ENOUGH_BYTES;
	
	if (isLE) {
		
		ret = buf->buf[buf->pos + 1];
		ret <<= 8;
		ret += buf->buf[buf->pos + 0];
	}
	else {
	
		ret = buf->buf[buf->pos + 0];
		ret <<= 8;
		ret += buf->buf[buf->pos + 1];
	}	
	
	if (ret < 0xd000 || ret >= 0xe000) {
	
		buf->pos += 2;
		return ret;
	}
	
	if (ret > 0xdc00) {	//unexpected second half of a surrogate pair
		
		buf->pos += 2;				//consume it
		return RW_ERR_INVAL_CHAR;
	}
	
	//we have a low surrogate pair - look for high
	if (buf->len - buf->pos < 4)
		return RW_ERR_NOT_ENOUGH_BYTES;
	
	if (isLE) {
		
		part2 = buf->buf[buf->pos + 3];
		part2 <<= 8;
		part2 += buf->buf[buf->pos + 2];
	}
	else {
	
		part2 = buf->buf[buf->pos + 2];
		part2 <<= 8;
		part2 += buf->buf[buf->pos + 3];
	}
	
	if (part2 < 0xdc00 || part2 >= 0xe000) {	//not second half of a surrogate pair - consume the first half and return
		
		buf->pos += 2;				//consume it
		return RW_ERR_INVAL_CHAR;
	}
	
	//we have a valid surrogate pair - decode it
	ret &= 0x03ff;
	part2 &= 0x03ff;
	
	return (((UInt32)ret) << 10) + part2;
}

static Int32 myTxtPrvReadUTF16LE(struct BufferState *buf)
{
	return myTxtPrvReadUTF16_generic(buf, true);
}

static Int32 myTxtPrvReadUTF16BE(struct BufferState *buf)
{
	return myTxtPrvReadUTF16_generic(buf, false);
}

static Int8 myTxtPrvWriteAscii(struct BufferState *buf, UInt32 chr)		//zero on error
{
	if (chr && (chr < 0x20 || chr >= 0x80))
		return RW_ERR_INVAL_CHAR;

	if (buf->buf) {
		if (buf->pos >= buf->len)
			return RW_ERR_NOT_ENOUGH_BYTES;
		
		buf->buf[buf->pos] = chr;
	}
	
	buf->pos++;
	return 0;
}

static Int8 myTxtPrvWritePalmLatin(struct BufferState *buf, UInt32 chr)		//zero on error
{
	UInt8 wri;
	
	if (chr && (chr < 0x20))
		return RW_ERR_INVAL_CHAR;
	else if (chr < 0x80 || chr >= 0xa0)
		wri = chr;
	else {
		for (wri = 0; wri < sizeof(mPalmLatinHighChars); wri++) {	//not common enough for me to care that it is slow
			if (mPalmLatinHighChars[wri] == (Int32)chr)
				break;
		}
		if (wri == sizeof(mPalmLatinHighChars))
			return RW_ERR_INVAL_CHAR;
		wri += 0x80;
	}
	
	if (buf->buf) {
		if (buf->pos >= buf->len)
			return RW_ERR_NOT_ENOUGH_BYTES;
		
		buf->buf[buf->pos] = wri;
	}
	
	buf->pos++;
	return 0;
}

static Int8 myTxtPrvWriteUCS2(struct BufferState *buf, UInt32 chr)		//zero on error
{
	UInt8 wri;
	
	if (chr >= 0x10000)
		return RW_ERR_INVAL_CHAR;
	
	if (buf->buf) {
		if (buf->len - buf->pos < 2)
			return RW_ERR_NOT_ENOUGH_BYTES;
		
		buf->buf[buf->pos + 0] = ((UInt16)chr) >> 8;
		buf->buf[buf->pos + 1] = chr;
	}
	
	buf->pos += 2;
	return 0;
}

static Int8 myTxtPrvWriteUTF8(struct BufferState *buf, UInt32 chr)		//zero on error
{
	if (chr < 0x80) {
		
		if (buf->buf) {
			
			if (buf->pos >= buf->len)
				return RW_ERR_NOT_ENOUGH_BYTES;
			
			buf->buf[buf->pos] = chr;
		}
		buf->pos++;
	}
	else if (chr < 0x800) {
		
		if (buf->buf) {
			
			if (buf->len - buf->pos < 2)
				return RW_ERR_NOT_ENOUGH_BYTES;
			
			buf->buf[buf->pos + 0] = 0xc0 + (((UInt16)chr) >> 6);
			buf->buf[buf->pos + 1] = 0x80 + (((UInt8)chr) & 0x3f);
		}
		buf->pos += 2;
	}
	else if (chr < 0x10000) {
		
		if (buf->buf) {
			
			if (buf->len - buf->pos < 3)
				return RW_ERR_NOT_ENOUGH_BYTES;
			
			buf->buf[buf->pos + 0] = 0xe0 + (((UInt16)chr) >> 12);
			buf->buf[buf->pos + 1] = 0x80 + ((((UInt16)chr) >> 6) & 0x3f);
			buf->buf[buf->pos + 2] = 0x80 + (((UInt8)chr) & 0x3f);
		}
		buf->pos += 3;
	}
	else if (chr < 0x110000) {
		
		if (buf->buf) {
			
			if (buf->len - buf->pos < 4)
				return RW_ERR_NOT_ENOUGH_BYTES;
			
			buf->buf[buf->pos + 0] = 0xf0 + (chr >> 18);
			buf->buf[buf->pos + 1] = 0x80 + ((chr >> 12) & 0x3f);
			buf->buf[buf->pos + 2] = 0x80 + ((((UInt16)chr) >> 6) & 0x3f);
			buf->buf[buf->pos + 3] = 0x80 + (((UInt8)chr) & 0x3f);
		}
		buf->pos += 4;
	}
	else
		return RW_ERR_INVAL_CHAR;
	
	return 0;
}

static Int8 myTxtPrvWriteUTF16_generic(struct BufferState *buf, UInt32 chr, Boolean isLE)
{
	if (chr >= 0xd800 && chr < 0xe000)
		return RW_ERR_INVAL_CHAR;
	else if (chr < 0x10000) {
		
		if (buf->buf) {
			if (buf->len - buf->pos < 2)
				return RW_ERR_NOT_ENOUGH_BYTES;
			
			if (isLE) {
				
				buf->buf[buf->pos + 1] = ((UInt16)chr) >> 8;
				buf->buf[buf->pos + 0] = chr;
			}
			else {
			
				buf->buf[buf->pos + 0] = ((UInt16)chr) >> 8;
				buf->buf[buf->pos + 1] = chr;
			}
		}
		
		buf->pos += 2;
	}
	else if (chr >= 0x110000)
		return RW_ERR_INVAL_CHAR;
	else {		//write surrogate pair
		
		if (buf->buf) {
			
			UInt16 p1 = 0xd800 + (((UInt16)chr) >> 10);
			UInt16 p2 = 0xdc00 + (((UInt16)chr) & 0x3ff);
			
			if (buf->len - buf->pos < 4)
				return RW_ERR_NOT_ENOUGH_BYTES;
			
			if (isLE) {
				
				buf->buf[buf->pos + 1] = p1 >> 8;
				buf->buf[buf->pos + 0] = p1;
				buf->buf[buf->pos + 3] = p2 >> 8;
				buf->buf[buf->pos + 2] = p2;
			}
			else {
			
				buf->buf[buf->pos + 0] = p1 >> 8;
				buf->buf[buf->pos + 1] = p1;
				buf->buf[buf->pos + 2] = p2 >> 8;
				buf->buf[buf->pos + 3] = p2;
			}
		}
		
		buf->pos += 4;
	}
	
	return 0;
}

static Int8 myTxtPrvWriteUTF16BE(struct BufferState *buf, UInt32 chr)		//zero on error
{
	return myTxtPrvWriteUTF16_generic(buf, chr, false);
}

static Int8 myTxtPrvWriteUTF16LE(struct BufferState *buf, UInt32 chr)		//zero on error
{
	return myTxtPrvWriteUTF16_generic(buf, chr, true);
}

Err MyTxtConvertEncoding(Boolean newConversion, TxtConvertStateType *state, UInt8 *srcTextP, UInt16 *ioSrcBytes, CharEncodingType srcEncoding, UInt8 *dstTextP, UInt16 *ioDstBytes, CharEncodingType dstEncoding, const Char *substitutionStr, UInt16 substitutionLen)
{
	Int8 (*writeF)(struct BufferState *buf, UInt32 chr) = NULL;
	Int32 (*readF)(struct BufferState *buf) = NULL;
	struct BufferState bufOut, bufIn;
	Err err = errNone;
		
	switch (srcEncoding) {
		case charEncodingAscii:		readF = myTxtPrvReadAscii;			break;
		case charEncodingPalmLatin:	readF = myTxtPrvReadPalmLatin;		break;
		case charEncodingUCS2:		readF = myTxtPrvReadUCS2;			break;
		case charEncodingUTF8:		readF = myTxtPrvReadUTF8;			break;
		case charEncodingUTF16BE:	readF = myTxtPrvReadUTF16BE;		break;
		case charEncodingUTF16LE:	readF = myTxtPrvReadUTF16LE;		break;
		default:					return txtErrUnknownEncoding;
	}
	
	switch (dstEncoding) {
		case charEncodingAscii:		writeF = myTxtPrvWriteAscii;		break;
		case charEncodingPalmLatin:	writeF = myTxtPrvWritePalmLatin;	break;
		case charEncodingUCS2:		writeF = myTxtPrvWriteUCS2;			break;
		case charEncodingUTF8:		writeF = myTxtPrvWriteUTF8;			break;
		case charEncodingUTF16BE:	writeF = myTxtPrvWriteUTF16BE;		break;
		case charEncodingUTF16LE:	writeF = myTxtPrvWriteUTF16LE;		break;
		default:					return txtErrUnknownEncoding;
	}	

	//verify param correctness (even the ones we do not use)
	if (!ioSrcBytes || !ioDstBytes || (substitutionLen && !substitutionStr) || (*ioDstBytes && !dstTextP) || (*ioSrcBytes && !srcTextP) || (!newConversion && !state))
		return sysErrParamErr;

	bufOut.buf = dstTextP;
	bufOut.len = *ioDstBytes;
	bufOut.pos = 0;
	bufIn.buf = srcTextP;
	bufIn.len = *ioSrcBytes;
	bufIn.pos = 0;

	while (bufIn.pos < bufIn.len) {
		
		struct BufferState bufTmp = bufIn;
		Int32 chr = readF(&bufIn);
		Int8 wriSta;
		
		if (chr == RW_ERR_NOT_ENOUGH_BYTES) {
			
			err = txtErrConvertUnderflow;
			bufIn = bufTmp;
			break;
		}
		
		if (chr == RW_ERR_INVAL_CHAR || ((wriSta = writeF(&bufOut, chr)) == RW_ERR_INVAL_CHAR)) {
		
			err = txtErrNoCharMapping;
			
			if (!substitutionStr || (bufOut.buf && bufOut.len - bufOut.pos < substitutionLen)) {		//no substitute strig or it will not fit?
				
				bufIn = bufTmp;		//rewind to the char in question
				break;
			}
			
			if (bufOut.buf)
				MemMove(bufOut.buf + bufOut.pos, substitutionStr, substitutionLen);
			bufOut.pos += substitutionLen;
		}
		else if (wriSta == RW_ERR_NOT_ENOUGH_BYTES) {
			
			err = txtErrConvertOverflow;
			bufIn = bufTmp;		//rewind to the char in question
			break;
		}
	}
	
	//write the zero that the bug allows
	if (err == errNone && writeF(&bufOut, 0) == RW_ERR_NOT_ENOUGH_BYTES)
		err = txtErrConvertOverflow;
	
	
	*ioSrcBytes = bufIn.pos;
	*ioDstBytes = bufOut.pos;
	
	return err;
}

Err MyTxtConvertEncodingV35(UInt8* srcTextP, UInt16* ioSrcBytes, CharEncodingType srcEncoding, UInt8* dstTextP, UInt16* ioDstBytes, CharEncodingType dstEncoding)
{
	return MyTxtConvertEncoding(true, NULL, srcTextP, ioSrcBytes, srcEncoding, dstTextP, ioDstBytes, dstEncoding, NULL, 0);
}

void MyIntlSelUnsupported(UInt16 sel, UInt32 from)
{
	char x[48];
	
	StrCopy(x, "Unsupported Intl selector ");
	StrIToA(x + 26, sel);
	StrCat(x, " from 0x");
	StrIToH(x + StrLen(x), from);
	SysFatalAlert(x);
	SysReset();
}

void MyIntlInit(void)
{
	//nothing
}

UInt8 MyTxtByteAttr(UInt8 byte)
{
	return byteAttrSingle;
}

UInt16 MyTxtCharAttr(WChar chr)
{
	if (chr >= 0xff80)
		chr &= 0xff;
	else if (chr >= 0x100)
		return 0;
	return GetCharAttr()[chr];
}

UInt16 MyTxtCharXAttr(WChar chr)
{
	return 0;
}

UInt16 MyTxtCharSize(WChar chr)
{
	return 1;
}

UInt16 MyTxtGetPreviousChar(const UInt8 *inText, UInt32 inOffset, WChar *outCharP)
{
	if (!inOffset) {
		
		if (outCharP)
			*outCharP = 0;
		
		return 0;
	}
	else {
		
		if (outCharP)
			*outCharP = inText[inOffset - 1];
		
		return 1;
	}
}

UInt16 MyTxtGetNextChar(const UInt8 *inText, UInt32 inOffset, WChar *outCharP)
{
	if (outCharP)
		*outCharP = inText[inOffset];
	
	return 1;
}

WChar MyTxtGetChar(const char *inText, UInt32 inOffset)
{
	return (UInt8)inText[inOffset];
}

UInt16 MyTxtSetNextChar(char *ioText, UInt32 inOffset, WChar inChar)
{
	ioText[inOffset] = inChar;
	
	return 1;
}

WChar MyTxtCharBounds(const UInt8 *inText, UInt32 inOffset, UInt32 *outStartP, UInt32 *outEndP)
{
	*outStartP = inOffset;
	*outEndP = inOffset + 1;
	
	return inText[inOffset];
}

//there are two possible forms (OS3.5+ and before-OS3.5). They differ in return type and params! We use assumptions about param order in 68k to account for this
struct TxtPrepFindStringParamsV1 {
	const UInt8* inSource;
	UInt8* outDest;
	UInt16 inDstSize;
};
struct TxtPrepFindStringParamsV2 {
	const UInt8* inSource;
	UInt16 inSrcLen;
	UInt8* outDest;
	UInt16 inDstSize;
};
UInt16 MyTxtPrepFindString(const UInt8 *inSourceP, ...)
{
	const UInt8 *caselessTable = GetCharCaselessValue();
	const UInt8 *inSource = inSourceP;
	UInt16 inDstSize, inSrcLen;
	UInt8 v, *outDest;
	UInt32 osVer;
	
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osVer) || osVer < sysMakeROMVersion(3, 5, 0, 0, 0)) {
		
		const struct TxtPrepFindStringParamsV1 *paramsV1 = (struct TxtPrepFindStringParamsV1*)&inSourceP;
		
		inSrcLen = StrLen((const char*)inSource);
		outDest = paramsV1->outDest;
		inDstSize = paramsV1->inDstSize;
	}
	else {
		
		const struct TxtPrepFindStringParamsV2 *paramsV2 = (struct TxtPrepFindStringParamsV2*)&inSourceP;
	
		inSrcLen = paramsV2->inSrcLen;
		outDest = paramsV2->outDest;
		inDstSize = paramsV2->inDstSize;
	}
	
	//reserve stace for terminator
	inDstSize--;
	
	//convert
	while (inSrcLen-- && inDstSize-- && (v = *inSource++) != 0)
		*outDest++ = caselessTable[v];
	
	//terminate
	*outDest = 0;
	
	return inSource - inSourceP;
}

static Boolean txtPrvIsPunctuation(const UInt16 *attrTable, UInt8 chr)
{
	return !!(attrTable[chr] & (_XS | _CN | _PU | _SP));
}

Boolean MyTxtFindString(const UInt8 *inSourceStr, const UInt8 *inTargetStr, UInt32 *outPosP, UInt16 *outLengthP)
{
	const UInt8 *caselessTable = GetCharCaselessValue(), *inSourceStrOrig = inSourceStr;
	const UInt16 *attrTable = GetCharAttr();
	UInt8 chr;
	
	//we are meant to only match at word start
	
	while ((chr = *inSourceStr++) != 0) {
		
		UInt16 i = 0;
		UInt8 diff;
		
		chr = caselessTable[chr];
		
		//match of first char?
		if (chr != *inTargetStr)
			continue;
		
		//word start == string start || prev_char is delimeter
		if (inSourceStr != inSourceStrOrig && !txtPrvIsPunctuation(attrTable, inSourceStr[-2]))
			continue;
		
		//we have a word start whose first char matches first search char. let's see if the rest matches...
		for (i = 0; (chr = inTargetStr[i]) && !(diff = inTargetStr[i] - caselessTable[inSourceStr[i]]); i++);
		
		if (!chr || !diff) {	//match
			
			*outPosP = inSourceStr - inSourceStrOrig;
			*outLengthP = i;
			return true;
		}
	}
	return false;
}

UInt16 MyTxtReplaceStr(UInt8 *ioStr, UInt16 inMaxLen, const UInt8 *inParamStr, UInt16 inParamNum)
{
	UInt16 nReplacements = 0, replLen = 0, curLen = StrLen((const char*)ioStr);
	UInt8 *lookAt = ioStr, *foundAt;
	
	if (inParamStr)
		replLen = StrLen((const char*)inParamStr);
	
	while ((foundAt = (UInt8*)StrChr((char*)lookAt, '^')) != NULL) {
		
		lookAt = foundAt + 1;
		
		if (foundAt[1] != '0' + inParamNum)
			continue;
		
		//match! replace if needed
		if (inParamStr) {
			
			if (curLen + replLen - 2 > inMaxLen)
				break;
			
			//it fits - replace
			MemMove(foundAt + replLen, foundAt + 2, ioStr + curLen - foundAt - 1);
			MemMove(foundAt, inParamStr, replLen);
			lookAt += replLen - 1;
			curLen += replLen - 2;
		}
		nReplacements++;
	}
	
	return nReplacements;
}

Boolean MyTxtWordBounds(const UInt8 *inText, UInt32 inLength, UInt32 inOffset, UInt32 *outStart, UInt32 *outEnd)
{
	const UInt16 *attrTable = GetCharAttr();
	Boolean wasAlphanum;
	UInt8 chr;
	UInt32 t;
	
	//As per disassembly, some punctuation chars (singleCloseCommaQuoteChr and chrApostrophe)
	// are treated as non-punctuation if between non-punctuation chars. Why is a mystery, but my
	// guess is that it is to deal with contractions.
	
	//sanity check
	if (!inLength || inOffset >= inLength) {
		
		*outStart = inLength;
		*outEnd = inLength;
		return false;
	}
	
	//what sort of thing is it?
	wasAlphanum = txtPrvIsPunctuation(attrTable, inText[inOffset]);
	
	//search for start
	t = inOffset;
	while (t) {
		if (txtPrvIsPunctuation(attrTable, chr = inText[t]) != wasAlphanum &&
				(!wasAlphanum || t <= 1 || (chr != singleCloseCommaQuoteChr && chr != chrApostrophe) || txtPrvIsPunctuation(attrTable, inText[t - 1])))
			break;
		t--;
	}
	*outStart = t;
		
	//search for end
	t = inOffset + 1;
	while (t < inLength) {
		if (txtPrvIsPunctuation(attrTable, chr = inText[t]) != wasAlphanum &&
				(!wasAlphanum || t == inLength - 1 || (chr != singleCloseCommaQuoteChr && chr != chrApostrophe) || txtPrvIsPunctuation(attrTable, inText[t + 1])))
			break;
		t++;
	}
	*outEnd = t;
	
	return wasAlphanum;
}

CharEncodingType MyTxtCharEncoding(WChar inChar)
{
	if (inChar >= 0xff80)
		inChar &= 0xff;
	
	if (inChar < 0x80)
		return charEncodingAscii;
	if (inChar < 0xa0)
		return charEncodingCP1252;
	if (inChar < 0x100)
		return charEncodingISO8859_1;
	return charEncodingUnknown;
}

CharEncodingType MyTxtStrEncoding(const UInt8 *inStr)
{
	Boolean haveISO8859_1 = false;
	UInt8 ch;
	
	while ((ch = *inStr++) != 0) {	//charEncodingUnknown case is impossible
	
		switch (MyTxtCharEncoding(ch)) {
			case charEncodingCP1252:
				return charEncodingCP1252;	//immediate
			
			case charEncodingISO8859_1:
				haveISO8859_1 = true;
				break;
		}
	}
	
	return haveISO8859_1 ? charEncodingISO8859_1 : charEncodingAscii;
}

const char* MyTxtEncodingName(CharEncodingType inEncoding)
{
	switch (inEncoding) {
		case charEncodingAscii:
			return encodingNameAscii;
		
		case charEncodingISO8859_1:
			return encodingNameISO8859_1;
		
		case charEncodingPalmLatin:
		case charEncodingCP1252:
			return encodingNameCP1252;
		
		default:
			return "";
	}
}

CharEncodingType MyTxtMaxEncoding(CharEncodingType a, CharEncodingType b)
{
	if (a == b)
		return a;
	
	if (a == charEncodingUnknown || b == charEncodingUnknown)
		return charEncodingUnknown;
	
	if (a == charEncodingAscii)
		return b;
	
	if (a == charEncodingISO8859_1)
		return (b == charEncodingCP1252 || b == charEncodingPalmLatin) ? b : charEncodingISO8859_1;
	
	if (a == charEncodingCP1252)
		return (b == charEncodingPalmLatin) ? charEncodingPalmLatin : charEncodingCP1252;
	
	return a;
}

static const UInt8 mLutToUpper[] = {	//min, max, adjustment
		0x61, 0x7A, 0xE0,
		0x9A, 0x9A, 0xF0,
		0x9C, 0x9C, 0xF0,
		0x9E, 0x9E, 0xF0,
		0xE0, 0xF6, 0xE0,
		0xF8, 0xFF, 0xE0,
	};

static const UInt8 mLutToLower[] = {	//min, max, adjustment
		0x41, 0x5A, 0x20,
		0x8A, 0x8A, 0x10,
		0x8C, 0x8C, 0x10,
		0x8E, 0x8E, 0x10,
		0x9F, 0x9F, 0x60,
		0xC0, 0xD6, 0x20,
		0xD8, 0xDE, 0x20,
		0x00, 0xFF, 0x00, 
	};

Err MyTxtTransliterate(const UInt8 *inSrcText, UInt16 inSrcLength, UInt8 *outDstText, UInt16 *ioDstLength, TranslitOpType inOp)
{
	const UInt8 *lutPtr;

	switch (inOp &~ translitOpPreprocess) {
		case translitOpUpperCase:	lutPtr = mLutToUpper;	break;
		case translitOpLowerCase:	lutPtr = mLutToLower;	break;
		default: return txtErrUknownTranslitOp;
	}
	
	if (inOp & translitOpPreprocess) {
	
		*ioDstLength = inSrcLength;
		return errNone;
	}
	
	if (*ioDstLength < inSrcLength)
		return txtErrTranslitOverflow;
	*ioDstLength = inSrcLength;
	
	//clever logic from glue lib :)
	while (inSrcLength--) {
		
		const UInt8 *lut = lutPtr;
		UInt8 chr = *inSrcText++;
		
		while (1) {
			if (chr < *lut++)
				break;
			if (chr > *lut++) {
				lut++;
				continue;
			}
			chr += *lut;
			break;
		}
		
		*outDstText = chr;
	}
	
	return errNone;
}

Boolean MyTxtCharIsValid(WChar inChar)
{
	return inChar < 0x100;
}

static Int16 txtPrvDoCompare(const UInt8 *s1, UInt16 s1Len, UInt16 *s1MatchLen, const UInt8 *s2, UInt16 s2Len, UInt16 *s2MatchLen, const UInt8* lut)
{
	UInt16 commonLen = s1Len < s2Len ? s2Len : s2Len, matchLen = 0;
	Int16 ret = 0;
	UInt8 ch;
	
	while (commonLen) {
		
		if (!(ch = *s1) || (ret = lut[ch] - lut[*s2]))
			break;
		
		s1++;
		s2++;
		commonLen--;
		matchLen++;
	}
	
	if (!commonLen) {	//common length exhausted
		
		if (s1Len < s2Len)
			ret = *s2 ? -1 : 0;
		else if (s1Len > s2Len)
			ret = *s1 ? 1 : 0;
		//else, ret is already zero
	}
	
	if (s1MatchLen)
		*s1MatchLen = matchLen;
	
	if (s2MatchLen)
		*s2MatchLen = matchLen;
	
	return ret;
}

Int16 MyTxtCompare(const UInt8 *s1, UInt16 s1Len, UInt16 *s1MatchLen, const UInt8 *s2, UInt16 s2Len, UInt16 *s2MatchLen)
{
	Int16 ret = txtPrvDoCompare(s1, s1Len, s1MatchLen, s2, s2Len, s2MatchLen, GetCharCaselessValue());
	
	if (!ret)
		ret = txtPrvDoCompare(s1, s1Len, s1MatchLen, s2, s2Len, s2MatchLen, GetCharSortValue());
	
	return ret;
}

Int16 MyTxtCaselessCompare(const UInt8 *s1, UInt16 s1Len, UInt16 *s1MatchLen, const UInt8 *s2, UInt16 s2Len, UInt16 *s2MatchLen)
{
	return txtPrvDoCompare(s1, s1Len, s1MatchLen, s2, s2Len, s2MatchLen, GetCharCaselessValue());
}

Boolean MyIntlHandleEvent(SysEventType *evt, Boolean wtf)
{
	return false;
}

Int16 MyTxtCharWidth(WChar inChar)
{
	if (inChar >= 0xff80)
		inChar &= 0xff;
	
	return FntCharWidth(inChar);
}

UInt32 MyTxtGetTruncationOffset(const UInt8 *inText, UInt32 inOffset)
{
	return inOffset;
}

//why this instead of the glue library? this is actually multybyte aware due to proper use of TxtReplaceStr() in the ROM
char* MyTxtParamString(const char* inTemplate, const char *param0, const char *param1, const char *param2, const char *param3)
{
	MemHandle retH = MemHandleNew(StrLen(inTemplate) + 1);
	const char **params = &param0;		//tee hee
	char *ret;
	UInt8 i;
	
	if (!retH)
		return NULL;
	
	StrCopy(ret = MemHandleLock(retH), inTemplate);
	
	for (i = 0; i < 4; i++) {
		
		const char *rpl = params[i];
		UInt16 numSubsts;
		
		if (!rpl)
			rpl = "";
		
		numSubsts = TxtReplaceStr(ret, 0xffff, NULL, i);
		if (numSubsts) {
			
			MemHandleUnlock(retH);
			if (errNone != MemHandleResize(retH, MemHandleSize(retH) + numSubsts * (StrLen(rpl) - 2))) {
				
				MemHandleFree(retH);
				return NULL;
			}
			
			(void)TxtReplaceStr(ret = MemHandleLock(retH), 0xffff, rpl, i);
		}
	}
	
	return ret;
}

#ifdef __GNUC__
	
	#define STRINGIFY2(x)	#x
	#define STRINGIFY(x)	STRINGIFY2(x)
	
	void myIntlDispatchFull(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncoding) ", %d2		\n\t"
			"	bgt		iIntlSelUnsupported								\n\t"
			"	lsl.w   #2, %d2											\n\t"
			"	jmp		2(%pc, %d2.w)									\n\t"
			"															\n\t"
			"jumptable:													\n\t"
			"	bra.w	MyIntlInit										\n\t"
			"	bra.w	MyTxtByteAttr									\n\t"
			"	bra.w	MyTxtCharAttr									\n\t"
			"	bra.w	MyTxtCharXAttr									\n\t"
			"	bra.w	MyTxtCharSize									\n\t"
			"	bra.w	MyTxtGetPreviousChar							\n\t"
			"	bra.w	MyTxtGetNextChar								\n\t"
			"	bra.w	MyTxtGetChar									\n\t"
			"	bra.w	MyTxtSetNextChar								\n\t"
			"	bra.w	MyTxtCharBounds									\n\t"
			"	bra.w	MyTxtPrepFindString								\n\t"
			"	bra.w	MyTxtFindString									\n\t"
			"	bra.w	MyTxtReplaceStr									\n\t"
			"	bra.w	MyTxtWordBounds									\n\t"
			"	bra.w	MyTxtCharEncoding								\n\t"
			"	bra.w	MyTxtStrEncoding								\n\t"
			"	bra.w	MyTxtEncodingName								\n\t"
			"	bra.w	MyTxtMaxEncoding								\n\t"
			"	bra.w	MyTxtTransliterate								\n\t"
			"	bra.w	MyTxtCharIsValid								\n\t"
			"	bra.w	MyTxtCompare									\n\t"
			"	bra.w	MyTxtCaselessCompare							\n\t"
			"	bra.w	MyTxtCharWidth									\n\t"
			"	bra.w	MyTxtGetTruncationOffset						\n\t"
			"	bra.w	MyIntlGetRoutineAddress							\n\t"
			"	bra.w	MyIntlHandleEvent								\n\t"
			"	bra.w	MyTxtParamString								\n\t"
			"	bra.w	MyTxtConvertEncodingV35							\n\t"
			"	bra.w	MyTxtConvertEncoding							\n\t"
			"															\n\t"
			"MyIntlGetRoutineAddress:									\n\t"	//uses the very table above
			"	move.w	4(%sp), %d0										\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncoding) ", %d0		\n\t"
			"	bgt.s	MyIntlGetRoutineAddress_invalid					\n\t"
			"	lsl.w   #2, %d0											\n\t"
			"	lea     jumptable + 2(%pc, %d0), %a0					\n\t"
			"	adda.w  (%a0), %a0										\n\t"
			"	rts														\n\t"
			"															\n\t"
			"MyIntlGetRoutineAddress_invalid:							\n\t"
			"	suba.l	%a0, %a0										\n\t"
			"	rts														\n\t"
			"															\n\t"
			"iIntlSelUnsupported:										\n\t"
			"	move.l  (%sp), -(%sp)									\n\t"
			"	move.w	%d2, -(%sp)										\n\t"
			"	bsr		MyIntlSelUnsupported							\n\t"
			"	lea     6(%sp), %sp										\n\t"
			"	rts														\n\t"
		);
	}
	
	void pIntlDispatch_pPsrInit(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	move.l  (%sp), %d0					\n\t"
			"	subi.l  #0xAAAAAAAA, %d0			\n\t"	//start addr of System.prc('boot', 10102)
			"	bmi.s   not_rom						\n\t"
			"	subi.l  #0xAAAAAAAA, %d0			\n\t"	//length of System.prc('boot', 10102)
			"	bpl.s   not_rom						\n\t"
			"	movea.l #0xAAAAAAAA, %a0			\n\t"	//old trap
			"	jmp     (%a0)						\n\t"
			"not_rom:								\n\t"
			"	bra     myIntlDispatchFull			\n\t"
		);
	}
	
	void myIntlDispatch_withTxtParamString(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	movea.l #0xaaaaaaaa, %a0									\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncodingV35) ", %d2		\n\t"
			"	beq     MyTxtConvertEncodingV35								\n\t"	//cannot jump directly cause CW relaxes jump in a way that leaves behind garbage. trying to give ".l" or ".s" suffix to "beq" does nothing
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncoding) ", %d2			\n\t"
			"	beq     MyTxtConvertEncoding								\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtParamString) ", %d2				\n\t"
			"	beq     MyTxtParamString									\n\t"
			"	cmpi.w	#" STRINGIFY(intlIntlGetRoutineAddress) ", %d2		\n\t"
			"	beq     MyIntlGetRoutineAddress2							\n\t"
			"	jmp     (%a0)												\n\t"
			"																\n\t"
			"MyIntlGetRoutineAddress2:										\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncodingV35) ", 4(%sp)	\n\t"
			"	beq     geta_TxtConvertEncodingV35							\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncoding) ", 4(%sp)		\n\t"
			"	beq     geta_MyTxtConvertEncoding							\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtParamString) ", 4(%sp)			\n\t"
			"	beq     geta_MyTxtParamString								\n\t"
			"	cmpi.w	#" STRINGIFY(intlIntlGetRoutineAddress) ", 4(%sp)	\n\t"
			"	beq     geta_IntlGetRoutineAddress							\n\t"
			"	jmp     (%a0)												\n\t"
			"																\n\t"
			"geta_TxtConvertEncodingV35:									\n\t"
			"	lea     MyTxtConvertEncodingV35(%pc), %a0					\n\t"
			"	rts															\n\t"
			"																\n\t"
			"geta_MyTxtConvertEncoding:										\n\t"
			"	lea     MyTxtConvertEncoding(%pc), %a0						\n\t"
			"	rts															\n\t"
			"																\n\t"
			"geta_MyTxtParamString:											\n\t"
			"	lea     MyTxtParamString(%pc), %a0							\n\t"
			"	rts															\n\t"
			"																\n\t"
			"geta_IntlGetRoutineAddress:									\n\t"
			"	lea     MyIntlGetRoutineAddress2(%pc), %a0					\n\t"
		);
	}
	
	void myIntlDispatch_noTxtParamString(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	movea.l #0xaaaaaaaa, %a0									\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncodingV35) ", %d2		\n\t"
			"	beq     MyTxtConvertEncodingV35								\n\t"	//cannot jump directly cause CW relaxes jump in a way that leaves behind garbage. trying to give ".l" or ".s" suffix to "beq" does nothing
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncoding) ", %d2			\n\t"
			"	beq     MyTxtConvertEncoding								\n\t"
			"	cmpi.w	#" STRINGIFY(intlIntlGetRoutineAddress) ", %d2		\n\t"
			"	beq     MyIntlGetRoutineAddress3							\n\t"
			"	jmp     (%a0)												\n\t"
			"																\n\t"
			"MyIntlGetRoutineAddress3:										\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncodingV35) ", 4(%sp)	\n\t"
			"	beq     geta_TxtConvertEncodingV35							\n\t"
			"	cmpi.w	#" STRINGIFY(intlTxtConvertEncoding) ", 4(%sp)		\n\t"
			"	beq     geta_MyTxtConvertEncoding							\n\t"
			"	cmpi.w	#" STRINGIFY(intlIntlGetRoutineAddress) ", 4(%sp)	\n\t"
			"	beq     geta_IntlGetRoutineAddress							\n\t"
			"	jmp     (%a0)												\n\t"
		);
	}
	
	
	
#else

	static asm void myIntlDispatchFull(void)
	{
		cmpi.w	#0 + intlTxtConvertEncoding, d2
		bgt		iIntlSelUnsupported
		lsl.w   #2, d2
		jmp		2(pc, d2.w)
	
	jumptable:
		jmp		MyIntlInit
		jmp		MyTxtByteAttr
		jmp		MyTxtCharAttr
		jmp		MyTxtCharXAttr
		jmp		MyTxtCharSize
		jmp		MyTxtGetPreviousChar
		jmp		MyTxtGetNextChar
		jmp		MyTxtGetChar
		jmp		MyTxtSetNextChar
		jmp		MyTxtCharBounds
		jmp		MyTxtPrepFindString
		jmp		MyTxtFindString
		jmp		MyTxtReplaceStr
		jmp		MyTxtWordBounds
		jmp		MyTxtCharEncoding
		jmp		MyTxtStrEncoding
		jmp		MyTxtEncodingName
		jmp		MyTxtMaxEncoding
		jmp		MyTxtTransliterate
		jmp		MyTxtCharIsValid
		jmp		MyTxtCompare
		jmp		MyTxtCaselessCompare
		jmp		MyTxtCharWidth
		jmp		MyTxtGetTruncationOffset
		jmp		MyIntlGetRoutineAddress
		jmp		MyIntlHandleEvent
		jmp		MyTxtParamString
		jmp		MyTxtConvertEncodingV35
		jmp		MyTxtConvertEncoding
		
		
	MyIntlGetRoutineAddress:			//uses the very table above
		move.w	4(sp),d0
		cmpi.w	#0 + intlTxtConvertEncoding, d0
		bgt		MyIntlGetRoutineAddress_invalid
		lea     jumptable, a0
		add.w   d0, d0
		add.w   d0, d0
		lea     2(a0, d0), a0
		adda.w  (a0), a0
		rts
		
	MyIntlGetRoutineAddress_invalid:
		suba.l	a0,a0
		rts
	
	iIntlSelUnsupported:
		move.l  (sp), -(sp)
		move.w	d2, -(sp)
		jsr		MyIntlSelUnsupported
		lea     6(sp), sp
		rts
	}
	
	asm static void pIntlDispatch_pPsrInit(void)
	{
		move.l  (sp), d0
		subi.l  #0xAAAAAAAA, d0	//start addr of System.prc('boot', 10102)
		bmi     not_rom
		subi.l  #0xAAAAAAAA, d0	//length of System.prc('boot', 10102)
		bpl     not_rom
		movea.l #0xAAAAAAAA, a0	//old trap
		jmp     (a0)
	not_rom:
		jmp     myIntlDispatchFull
	}
	
	static asm void myIntlDispatch_withTxtParamString(void)
	{
		movea.l #0xaaaaaaaa, a0
		cmpi.w	#0 + intlTxtConvertEncodingV35, d2
		beq     do_MyTxtConvertEncodingV35			//cannot jump directly cause CW relaxes jump in a way that leaves behind garbage. trying to give ".l" or ".s" suffix to "beq" does nothing
		cmpi.w	#0 + intlTxtConvertEncoding, d2
		beq     do_MyTxtConvertEncoding
		cmpi.w	#0 + intlTxtParamString, d2
		beq     do_MyTxtParamString
		cmpi.w	#0 + intlIntlGetRoutineAddress, d2
		beq     MyIntlGetRoutineAddress
		jmp     (a0)
	
	do_MyTxtConvertEncodingV35:
		jmp     MyTxtConvertEncodingV35
	
	do_MyTxtConvertEncoding:
		jmp     MyTxtConvertEncoding
	
	do_MyTxtParamString:
		jmp     MyTxtParamString
	
	MyIntlGetRoutineAddress:			//uses the very table above
		cmpi.w	#0 + intlTxtConvertEncodingV35, 4(sp)
		beq     geta_TxtConvertEncodingV35
		cmpi.w	#0 + intlTxtConvertEncoding, 4(sp)
		beq     geta_MyTxtConvertEncoding
		cmpi.w	#0 + intlTxtParamString, 4(sp)
		beq     geta_MyTxtParamString
		cmpi.w	#0 + intlIntlGetRoutineAddress, 4(sp)
		beq     geta_IntlGetRoutineAddress
		jmp     (a0)
	
	geta_TxtConvertEncodingV35:
		lea     MyTxtConvertEncodingV35, a0
		rts
	
	geta_MyTxtConvertEncoding:
		lea     MyTxtConvertEncoding, a0
		rts
	
	geta_MyTxtParamString:
		lea     MyTxtParamString, a0
		rts
	
	geta_IntlGetRoutineAddress:
		lea     MyIntlGetRoutineAddress, a0
		rts
	}
	
	static asm void myIntlDispatch_noTxtParamString(void)
	{
		movea.l #0xaaaaaaaa, a0
		cmpi.w	#0 + intlTxtConvertEncodingV35, d2
		beq     do_MyTxtConvertEncodingV35			//cannot jump directly cause CW relaxes jump in a way that leaves behind garbage. trying to give ".l" or ".s" suffix to "beq" does nothing
		cmpi.w	#0 + intlTxtConvertEncoding, d2
		beq     do_MyTxtConvertEncoding
		cmpi.w	#0 + intlIntlGetRoutineAddress, d2
		beq     MyIntlGetRoutineAddress
		jmp     (a0)
	
	do_MyTxtConvertEncodingV35:
		jmp     MyTxtConvertEncodingV35
	
	do_MyTxtConvertEncoding:
		jmp     MyTxtConvertEncoding
	
	MyIntlGetRoutineAddress:
		cmpi.w	#0 + intlTxtConvertEncodingV35, 4(sp)
		beq     geta_TxtConvertEncodingV35
		cmpi.w	#0 + intlTxtConvertEncoding, 4(sp)
		beq     geta_MyTxtConvertEncoding
		cmpi.w	#0 + intlIntlGetRoutineAddress, 4(sp)
		beq     geta_IntlGetRoutineAddress
		jmp     (a0)
	
	geta_TxtConvertEncodingV35:
		lea     MyTxtConvertEncodingV35, a0
		rts
	
	geta_MyTxtConvertEncoding:
		lea     MyTxtConvertEncoding, a0
		rts
	
	geta_IntlGetRoutineAddress:
		lea     MyIntlGetRoutineAddress, a0
		rts
	}
#endif

void __Startup__(void)
{
	UInt32 dummy, osVer;
	
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osVer))
		return;
	
	//if international feature set does not exist, we need to provide it entirely. This will be OS versions 3.0 and below
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumIntlMgr, &dummy)) {
			
		if (osVer < sysMakeROMVersion(2, 0, 0, 0, 0)) {
			
			//PalmOS 1.0 only:
			// We want to allow existing code to use the old code allocated to sysTrapIntlDispatch.
			// That would be sysTrapPsrInit. Luckily, the only code that does is in System.pc, ('boot', 10102) resource
			// we check for return address in there to decide what to do.
			// I hope nobody else patches this func before or after, else this logic will break
		
			DmOpenRef db = DmOpenDatabaseByTypeCreator(sysFileTSystem, sysFileCSystem, dmModeReadOnly);
			MemHandle mh = DmGet1Resource(sysResTBootCode, 10102);
			UInt32 sysCodeSize = MemHandleSize(mh);
			void *sysCodePtr = MemHandleLock(mh);
			
			MemSemaphoreReserve(true);
			*(UInt32*)(((UInt16*)pIntlDispatch_pPsrInit) + 2) = (UInt32)sysCodePtr;
			*(UInt32*)(((UInt16*)pIntlDispatch_pPsrInit) + 6) = sysCodeSize;
			*(UInt32*)(((UInt16*)pIntlDispatch_pPsrInit) + 10) = (UInt32)SysGetTrapAddress(sysTrapIntlDispatch);
			MemSemaphoreRelease(true);
			
			MemHandleUnlock(mh);
			DmReleaseResource(mh);
			DmCloseDatabase(db);
			
			SysSetTrapAddress(sysTrapIntlDispatch, pIntlDispatch_pPsrInit);
		}
		else
			SysSetTrapAddress(sysTrapIntlDispatch, myIntlDispatchFull);
		
		FtrSet(sysFtrCreator, sysFtrNumIntlMgr, 1);
	}
	else {
		void *patchF = NULL;
		
		//OS3.1 - 3.3 provide some of the IntlMgr but are missing TxtParamString and encoding conversion, so we provide just that
		//OS3.5 is missing just the encoding conversion
		//we do not want to override rom funcs as they may be better than ours, so we provide only what is missing
		
		if (!IntlGetRoutineAddress(intlTxtParamString))
			patchF = myIntlDispatch_withTxtParamString;
		else if (!IntlGetRoutineAddress(intlTxtConvertEncoding))
			patchF = myIntlDispatch_noTxtParamString;
		else
			return;
		
		MemSemaphoreReserve(true);
		*(volatile void**)(((UInt16*)patchF) + 1) = SysGetTrapAddress(sysTrapIntlDispatch);
		MemSemaphoreRelease(true);
		
		SysSetTrapAddress(sysTrapIntlDispatch, patchF);
	}
	
	MemHandleLock(DmGet1Resource('extn', 13));
}
