#ifndef _HOTSYNC_H_
#define _HOTSYNC_H_

#include <stdbool.h>
#include <stdint.h>

#define HOTSYNC_MIN_BUF_SIZE	128

extern uint8_t mTxRxBuf[];		//must be at least HOTSYNC_MIN_BUF_SIZE bytes big

struct HsDateTime {
	uint16_t year;
	uint8_t month, day, hr, min, sec, rfu;
} __attribute__((packed));

struct HsUserInfo {
	uint32_t id1;
	uint32_t id2;
	uint32_t lstSyncPcId;
	struct HsDateTime lastSyncSuccessTime, lastSyncTryTime;
	uint8_t userNameLen;
	uint8_t encPasswdLen;
	uint8_t userNameAndPasswd[];	//[userNameLen + encPasswdLen]
} __attribute__((packed));

struct HsSysInfo {
	uint32_t romVer;
	uint32_t locVer;
	uint8_t rfu;
	uint8_t prodIdLen;
	uint8_t prodId[];				//[prodIdLen]
} __attribute__((packed));

struct HsDbCreateReq {
	uint32_t crid;
	uint32_t type;
	uint8_t cardNo;
	uint8_t rfu;
	uint16_t dbFlags;
	uint16_t dbVersion;
	uint8_t dbName[32];		//zero terminated. can be var-len but we do not bother
} __attribute__((packed));

struct HsDbInfo {
	uint8_t structSz;				//size of this actual struct incl up to 2 bytes of null-padding
	uint8_t flags;					//only in later versions
	uint16_t dbFlags;
	uint32_t type;
	uint32_t crid;
	uint16_t dbVer;
	uint32_t dbMod;
	struct HsDateTime crTime;
	struct HsDateTime modTime;
	struct HsDateTime bkpTime;
	uint16_t thisIdx;				//index of this one
	char dbName[];
} __attribute__((packed));

struct HsCardInfo {
	uint8_t structSz;				//size of this struct (always multiple of2)
	uint8_t cardNo;
	uint16_t cardVer;				//as per card header
	struct HsDateTime crTime;		//as per card header
	uint32_t romSize;
	uint32_t ramSize;
	uint32_t ramFree;
	uint8_t cardNameSz;
	uint8_t cardManufSz;
	char cardNameAndManuf[];
} __attribute__((packed));

#define OPEN_MODE_READ				0x80
#define OPEN_MODE_WRITE				0x40
#define OPEN_MODE_SHOW_SECRET		0x10

struct HsRpcParam {
	char boolPassByRef;
	uint8_t size;
	//data here, multiple of 2 bytes
} __attribute__((packed));

struct HsRpcPkt {
	uint8_t val_0x2d;
	uint8_t val_0x00;
	uint16_t rfu;
	uint16_t trapNo;
	uint32_t retD0;		//integer reply
	uint32_t retA0;		//pointer reply (arguably not useful)
	uint16_t nParams;
	//struct HsRpcParam params[] here. IN REVERSE ORDER!!!!!
} __attribute__((packed));



enum HsSyncEndCause {
	HsSyncEndCauseNormal,
	HsSyncEndCauseOOM,
	HsSyncEndCauseUserReqd,
	HsSyncEndCauseOther,
	//later versions added more but 1.0 will not support them, so who cares?
};

enum HotsyncRecReadRet {
	HotsyncRecReadError,
	HotsyncRecExists,			//but may be zero-sized
	HotsyncRecDoesntExist,
};


//basic
uint32_t hotsyncInit(const void *incomingPkt, uint_fast16_t len, uint32_t desiredBaudrate);		//returns baudrate to use or zero on error
uint16_t hotsyncGetUserInfo(struct HsUserInfo *info, uint16_t memLen);					//returns num bytes written. will truncate to fit
uint16_t hotsyncGetSysInfo(struct HsSysInfo *info, uint16_t memLen);					//returns num bytes written. will truncate to fit
typedef void (*HsCardInfoCbk)(void* userData, const struct HsCardInfo *info);
bool hotsyncGetCardInfo(uint8_t cardNo, HsCardInfoCbk cbkF, void *cbkD);

//DBs
typedef bool (*HotsyncDbsEnumF)(void *userData, const struct HsDbInfo *info);	//return false to stop enum
bool hotsyncDbsEnum(uint8_t cardNo, bool includeROM, bool includeRAM, HotsyncDbsEnumF cbkF, void *cbkD);
int16_t hotsyncDbCreate(struct HsDbCreateReq *req);										//returns DB handle or negative on error
int16_t hotsyncDbOpen(uint8_t cardNo, const char *name, uint8_t mode);					//OPEN_MODE_*, returns DB handle or negative on error
bool hotsyncDbClose(uint8_t dbHandle);
bool hotsyncDbDelete(uint8_t cardNo, const char *name);

//appinfo/sortinfo blocks
bool hotsyncInfoBlockRead(uint8_t dbHandle, bool appInfo, uint16_t offset, uint16_t len, void *dst, uint16_t *fullSizeP);	//ask for too much, get garbge or error...
bool hotsyncInfoBlockWriteBegin(uint8_t dbHandle, bool appInfo, uint16_t size);
bool hotsyncInfoBlockWriteContinue(const void *data, uint16_t len);
bool hotsyncInfoBlockWriteFinish(bool appInfo);

//records
enum HotsyncRecReadRet hotsyncRecRead(uint8_t dbHandle, uint16_t idx, uint16_t offset, uint16_t len, void *dst, uint32_t *recIdP, uint16_t *recSizeP, uint8_t *recAttrP, uint8_t *categoryP);	//ask for too much, get garbge or error...
bool hotsyncRecWriteBegin(uint8_t dbHandle, uint16_t size, uint32_t recID, uint8_t recAttrs, uint8_t category);
bool hotsyncRecWriteContinue(const void *data, uint16_t len);
bool hotsyncRecWriteFinish(uint32_t *recIDP);
bool hotsyncRecDelete(uint8_t dbHandle, uint32_t recID);

//resources
bool hotsyncResRead(uint8_t dbHandle, uint16_t idx, uint16_t offset, uint16_t len, void *dst, uint32_t *resTypeP, uint16_t *resIDP, uint16_t *resSizeP);	//ask for too much, get garbge or error...
bool hotsyncResWriteBegin(uint8_t dbHandle, uint32_t resType, uint16_t resID, uint16_t len);
bool hotsyncResWriteContinue(const void *data, uint16_t len);
bool hotsyncResWriteFinish(void);
bool hotsyncResDelete(uint8_t dbHandle, uint32_t resType, uint16_t resID);

//very low-level RPC
bool hotsyncRPC(struct HsRpcPkt *pkt, uint16_t len);

//higher level constructs
bool hotsyncRpcFtrGet(uint32_t ftrCrid, uint16_t ftrId, uint32_t *ftrValP, uint16_t *errP);
bool hotsyncRpcPrefSetAppPreferencesV10(uint32_t crid, int_fast16_t version, const void *data, uint_fast16_t dataSz);
bool hotsyncRpcPrefGetAppPreferencesV10(uint32_t crid, int_fast16_t version, void *data, uint_fast16_t dataSz, bool *retValP);
int16_t hotsyncRpcMemReadByte(uint32_t addr);

//management
bool hotsyncOpenConduit(void);															//to be done before every conduit. really just changes UI to "synchrnizing" from "identifying user"
bool hotsyncEndSync(enum HsSyncEndCause cause);
bool hotsyncDemandResetAtEnd(void);


#endif
