#include <string.h>
#include <stdio.h>
#include "lpc802.h"
#include "../Src/proto.h"
#include "timebase.h"
#include "hotsync.h"
#include "pinmux.h"
#include "printf.h"
#include "swap.h"
#include "frag.h"
#include "uart.h"
#include "slk.h"
#include "fat.h"
#include "prc.h"
#include "sd.h"

#define OUR_BAUDRATE_MAX	(CLOCKRATE / 16)

#define MY_HW_VER			1
#define MY_FW_VER			0x020201ul	//24 bit only!

struct FroApi {
	uint32_t rfu[2];
	void (*set_fro_frequency)(uint32_t iFreq);
};

struct DivApi {
	union {
		struct {
			int32_t (*sidiv)(int32_t numerator, int32_t denominator);		//remainder in r1
			uint32_t (*uidiv)(uint32_t numerator, uint32_t denominator);		//remainder in r1
		};
		struct {
			uint64_t (*sidiv_full)(int32_t numerator, int32_t denominator);		//remainder in r1
			uint64_t (*uidiv_full)(uint32_t numerator, uint32_t denominator);		//remainder in r1
		};
	};
	void (*sidivmod)(int32_t *outP, int32_t numerator, int32_t denominator);
	void (*uidivmod)(uint32_t *outP, uint32_t numerator, uint32_t denominator);
};

struct RomApi {
	uint32_t rfu[3];
	struct FroApi *fro;
	struct DivApi *div;
};

typedef void (*IapApi)(const uint32_t *params, uint32_t *results);

#define ROM			(*(const struct RomApi**)0x0f001ff8)
#define IAP			((IapApi)0x0F001FF1)

#define SLK_TIME_SCALE			(17)







uint8_t mTxRxBuf[528];		//must be at least 512 bytes as it is used by FAT too
static uint8_t mHsDataBuffer[128];

static void tryHotsync(void *pkt, uint16_t len);
static void parseAndReplyToSasdPacket(struct SlkHdr *rxSlkHdr, uint16_t len, struct FragAckInfo *fi, bool isDupe);

void prPutchar(char chr)
{
	extern volatile uint32_t __mailbox;
	
	while (__mailbox & 0x80000000);
	__mailbox = 0x80000000 | (uint8_t)chr;
}

uint32_t slkExtGetCurTime(void)
{
	return timebaseGet() >> SLK_TIME_SCALE;
}

uint16_t slkExtRx(void *slkState, void *dstP, uint16_t nBytes, uint32_t endTime)
{
	int32_t duration = endTime - slkExtGetCurTime();

	//convert time to timebase time
	if (duration < 0)
		duration = 0;

	return uartRx(dstP, nBytes, timebaseGet() + ((((uint64_t)(uint32_t)duration)) << SLK_TIME_SCALE));
}

void slkExtTx(void *slkState, const void *srcP, uint16_t len)
{
	uartTx(srcP, len);
}

static bool hsPrvPreinst();

static uint_fast16_t getBattVoltage(void)
{
	uint32_t result, acc = 0;
	uint_fast8_t i;
	
	SYSCON->PDRUNCFG &=~ SYSCON_PDRUNCFG_ADC_PD_MASK;				//ADC powered on
	SYSCON->SYSAHBCLKCTRL0 |= SYSCON_SYSAHBCLKCTRL0_ADC_MASK;		//ADC clocked
	
	ADC0->CTRL = (7 << ADC_CTRL_CLKDIV_SHIFT);
	ADC0->SEQ_CTRL[0] = ADC_SEQ_CTRL_SEQ_ENA_MASK | ADC_SEQ_CTRL_TRIGPOL_MASK | (1 << (ADC_SEQ_CTRL_CHANNELS_SHIFT + ADC_NO_BATT));

	for (i = 0; i < 16; i++) {
		ADC0->SEQ_CTRL[0] |= ADC_SEQ_CTRL_START_MASK;
		while (!((result = ADC0->SEQ_GDAT[0]) & ADC_SEQ_GDAT_DATAVALID_MASK));
		result = (result & ADC_SEQ_GDAT_RESULT_MASK) >> ADC_SEQ_GDAT_RESULT_SHIFT;
		acc += result;
	}
	
	SYSCON->SYSAHBCLKCTRL0 &=~ SYSCON_SYSAHBCLKCTRL0_ADC_MASK;		//ADC clocked down
	SYSCON->PDRUNCFG |= SYSCON_PDRUNCFG_ADC_PD_MASK;				//ADC powered off
	
	return acc * 3300 / 65536;
}

static uint_fast8_t battVoltsToPercent(uint_fast16_t mv)
{
	static const uint16_t mVoltages[] = {800, 1050, 1120, 1170, 1205, 1225, 1265, 1295, 1320, 1400, 1500};
	uint_fast8_t i;
	
	for (i = 0; i < sizeof(mVoltages) / sizeof(*mVoltages); i++) {
		
		if (mv == mVoltages[i])
			return i * 10;
		if (mv < mVoltages[i])
			break;
	}
	//underflow?
	if (i == 0)
		return 0;
	//overlow?
	if (i == sizeof(mVoltages) / sizeof(*mVoltages))
		return 100;
	
	//scale
	return ((i - 1) * 10 * (mVoltages[i] - mv) + i * 10 * (mv - mVoltages[i - 1]) + (mVoltages[i] - mVoltages[i - 1]) / 2) / (mVoltages[i] - mVoltages[i - 1]);
}

void main(void)
{
	struct SlkHdr rxSlkHdr;
	uint16_t rxLen;
	uint8_t prevTid = 0;
	bool inited;
	
	ROM->fro->set_fro_frequency(30000);		//15MHz speed
	
	//system
	SYSCON->SYSAHBCLKCTRL0 |=
		SYSCON_SYSAHBCLKCTRL0_GPIO0_MASK |
		SYSCON_SYSAHBCLKCTRL0_SWM_MASK |
		SYSCON_SYSAHBCLKCTRL0_SPI0_MASK |
		SYSCON_SYSAHBCLKCTRL0_UART0_MASK |
		SYSCON_SYSAHBCLKCTRL0_IOCON_MASK |
		SYSCON_SYSAHBCLKCTRL0_CRC_MASK;
	SYSCON->IRQLATENCY = 0;										//no need to slow us down
	SWM0->PINENABLE0 = ~(0x30 | (1 << (10 + ADC_NO_BATT)));		//SWD pins are special, so is our ADC pin
	
	//LED
	IOCON->PIO[mIoconRegOrder[PIN_NO_LED]] = 0x20;
	GPIO->SET[0] = 1 << PIN_NO_LED;
	GPIO->DIRSET[0] = 1 << PIN_NO_LED;
	
	//battery adc pin
	IOCON->PIO[mIoconRegOrder[PIN_NO_BATT_ADC]] = 0x00;
	GPIO->DIRCLR[0] = 1 << PIN_NO_BATT_ADC;
	
	//CRC
	CRC->MODE = 0;
	
	asm volatile("cpsie i");
	
	timebaseInit();
	
	//uart
	uartInit();
	sdHwInit();
		
	//SD
	printf("trying sd\n");
	inited = sdCardInit();
	printf("init = %d\n", inited);
	printf("num sec: %d\n", sdGetNumSecs());
	
	//go
	while(1) {
		
		struct FragAckInfo fi;
		
		rxLen = sizeof(mTxRxBuf);
		
		if (uartSeenFrameErrs()) {	//maybe go back to 9600 baud
			printf("resetting due to frame errors\n");
			
			uartResetFrameErrors();
			uartSetBaudrate(9600);
		}
		
		if (!fragRecv(NULL, &rxSlkHdr, mTxRxBuf, &rxLen, slkExtGetCurTime() + 10000, false, &fi))	//no ack here yes
			continue;
		
		if (rxSlkHdr.pktTyp == PKT_TYP_HOTSYNC && rxSlkHdr.dstSlk == PKT_ADDR_DESKTOP_LINK && rxSlkHdr.srcSck == PKT_ADDR_DESKTOP_LINK) {
			
			(void)fragAck(NULL, &rxSlkHdr, &fi, NULL, 0);	//ack it
			
			tryHotsync(mTxRxBuf, rxLen);
			uartTxWait();
			timebaseDelay(12000000 / 1000);
			uartSetBaudrate(9600);	//in case hotsync changed it...
		}
		else if (rxSlkHdr.pktTyp == PKT_TYP_SASD && rxSlkHdr.dstSlk == PKT_ADDR_SASD && rxSlkHdr.srcSck == PKT_ADDR_SASD) {
		
			parseAndReplyToSasdPacket(&rxSlkHdr, rxLen, &fi, prevTid == rxSlkHdr.tid);
			prevTid = rxSlkHdr.tid;
		}
	}
	
	while(1);	//gStop gcc warning...
}

#ifdef ENABLE_PRINTF_LOGGING

	void __attribute__((used)) report_hard_fault(uint32_t* regs, uint32_t ret_lr, uint32_t *user_sr)
	{
		uint32_t *push = (ret_lr == 0xFFFFFFFD) ? user_sr : (regs + 8);
		
		printf("============ HARD FAULT ============\n");
		printf("R0  = 0x%08X    R8  = 0x%08X\n", push[0], regs[0]);
		printf("R1  = 0x%08X    R9  = 0x%08X\n", push[1], regs[1]);
		printf("R2  = 0x%08X    R10 = 0x%08X\n", push[2], regs[2]);
		printf("R3  = 0x%08X    R11 = 0x%08X\n", push[3], regs[3]);
		printf("R4  = 0x%08X    R12 = 0x%08X\n", regs[4], push[4]);
		printf("R5  = 0x%08X    SP  = 0x%08X\n", regs[5], push + 8);
		printf("R6  = 0x%08X    LR  = 0x%08X\n", regs[6], push[5]);
		printf("R7  = 0x%08X    PC  = 0x%08X\n", regs[7], push[6]);
		printf("RA  = 0x%08X    SR  = 0x%08X\n", ret_lr,  push[7]);
		
	}
	
	void __attribute__((used, noreturn, naked, noinline)) HardFault_Handler(void)
	{
		asm volatile(
				"push {r4-r7}			\n\t"
				"mov  r0, r8			\n\t"
				"mov  r1, r9			\n\t"
				"mov  r2, r10			\n\t"
				"mov  r3, r11			\n\t"
				"push {r0-r3}			\n\t"
				"mov  r0, sp			\n\t"
				"mov  r1, lr			\n\t"
				"mrs  r2, PSP			\n\t"
				"bl   report_hard_fault		\n\t"
				"1:				\n\t"
				"wfi				\n\t"
				"b 1b				\n\t"	//loop forever
				:::"memory");
	}

#endif





enum SdOpInProgress {
	SdNoneInProgress,
	SdReadInProgress,
	SdWriteInProgress,
};

struct SdState {
	enum SdOpInProgress curOp;
	uint32_t firstSec, numSec, doneSec;
};

struct SdState mState = {.curOp = SdNoneInProgress, };

static void sasdPrvSendErrorAck(struct SlkHdr *rxSlkHdr, struct FragAckInfo *fi, uint8_t errorType)
{
	struct SdErrorData ed = {.cmd = RSP_ERROR, .errorType = errorType, };
	
	if (!fragAck(NULL, rxSlkHdr, fi, &ed, sizeof(struct SdErrorData)))
		printf("cannot send err-ack\n");
}

static void selfErase(void)
{
	uint32_t params[4], results[3];
	
	params[0] = 50;
	params[1] = 0;
	params[2] = 15;
	printf("erasing\n");
	asm volatile("cpsid i");
	IAP(params, results);	//prepare
	printf("ret = %d\n", results[0]);
	params[0] = 59;
	params[1] = 0;
	params[2] = 0;
	params[3] = 0;
	IAP(params, results);
	printf("ret = %d\n", results[0]);
	
	SCB->AIRCR = 0x05FA0004;
	while(1);
}

static void parseAndReplyToSasdPacket(struct SlkHdr *rxSlkHdr, uint16_t len, struct FragAckInfo *fi, bool isDupe)
{
	static const uint32_t cmdIspEnter[] = {57};
	struct SdReadWriteReq *rwr = (struct SdReadWriteReq*)mTxRxBuf;
	struct SdBatteryData *batt = (struct SdBatteryData*)mTxRxBuf;
	struct SdInfoData *info = (struct SdInfoData*)mTxRxBuf;
	struct SdBaudData *baud = (struct SdBaudData*)mTxRxBuf;
	uint32_t secStart, secNum, secNow;
	uint8_t cmd, ok = RSP_OK;
	uint_fast16_t battVolts;
	bool isWrite = false;
	union SdFlags flags;
	
	if (len < 1)
		return;
	
	printf("cmd 0x%02x len %u\n", mTxRxBuf[0], len);
	
	switch (mTxRxBuf[0]) {
		case CMD_PING:					//send ack with pong reply
			if (len != sizeof(struct SdPingData))
				return;
			mTxRxBuf[0] = RSP_PONG;
			
			if (!fragAck(NULL, rxSlkHdr, fi, mTxRxBuf, sizeof(struct SdPingData)))
				printf("cannot send pong-ack\n");

			break;
		
		case CMD_GET_INFO:				//send ack with info reply
		
			if (len != 1)
				return;
			
			flags.value = sdGetFlags();
			
			memset(info, 0, sizeof(*info));
			info->cmd = RSP_INFO;
			
			if (flags.inited) {
			
				uint32_t snum;
				uint16_t oem;
				
				info->flags = SD_FLAG_COMMS | SD_FLAG_HAVE_CARD |
					(flags.v2 ? SD_FLAG_IS_NEW_VER : 0) |
					(flags.HC ? SD_FLAG_IS_HIGH_CAP : 0) |
					(flags.SD ? 0 : SD_FLAG_IS_MMC) |
					(flags.RO ? SD_FLAG_READ_ONLY : 0) |
					(flags.cmd23supported ? SD_FLAG_HAS_CMD_23 : 0) |
					(flags.hasDiscard ? SD_FLAG_HAS_DISCARD : 0);
					
				info->numBlocks = TO_U32(sdGetNumSecs());
				sdGetInfo(&info->manuf, &oem, &snum);
				info->oem = TO_U16(oem);
				info->snum = TO_U32(snum);
				info->flags = TO_U32(info->flags);
			}
			else
				info->flags = TO_U32(SD_FLAG_COMMS);
			
			info->ver.hwVers = MY_HW_VER;
			info->ver.fwVers[0] = (MY_FW_VER >> 16) & 0xff;
			info->ver.fwVers[1] = (MY_FW_VER >> 8) & 0xff;
			info->ver.fwVers[2] = MY_FW_VER & 0xff;
				
			if (!fragAck(NULL, rxSlkHdr, fi, info, sizeof(struct SdInfoData)))
				printf("cannot send info-ack\n");

			break;
		
		case CMD_WRITE:
			isWrite = true;
			//fallthrough
		case CMD_READ:
			if (len != sizeof(struct SdReadWriteReq))
				return;
			
			secStart = TO_U32(rwr->sector);
			secNum = TO_U32(rwr->numSec);
			
	//		printf("range %u + %u\n", (unsigned)secStart, (unsigned)secNum);
	
			if (secStart >= sdGetNumSecs() || !secNum || sdGetNumSecs() - secStart < secNum) {
				printf("out of range\n");
				return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_BOUNDS);
			}
			
			switch (mState.curOp) {
				case SdNoneInProgress:
					//all good
					break;
				
				case SdReadInProgress:
				case SdWriteInProgress:
				
					if (mState.numSec && mState.doneSec && mState.numSec != mState.doneSec) {	//if something was started but not yet ended, it is a sequence error
						printf("sequence error\n");
						return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_SEQUENCE);
					}
					break;
				
				default:
					return;
			}
			
			mState.firstSec = secStart;
			mState.numSec = secNum;
			mState.doneSec = 0;
			mState.curOp = isWrite ? SdWriteInProgress : SdReadInProgress;
			
			//send an empty ack
			if (!fragAck(NULL, rxSlkHdr, fi, &ok, sizeof(ok)))
				printf("cannot send ok-ack\n");
			
			return;
		
		case NOTICE_CTS:
			if (len != 1)
				return;
			
			//verify we are in read mode
			if (mState.curOp != SdReadInProgress)
				return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_STATE);
			
			//if not dupe, account for and send it
			if (!isDupe) {
				
				if (mState.doneSec >= mState.numSec)
					return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_SEQ2);
				
				secNow = mState.doneSec++;
			}
			else {
				
				if (!mState.doneSec)
					return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_SEQ2);
				
				secNow = mState.doneSec - 1;
			}
			
			mTxRxBuf[1] = NOTICE_XFER;
			
			secNow += mState.firstSec;
			//printf("**RD %u\n", secNow);

			if (!sdSecRead(secNow, mTxRxBuf + 2)) { /* align to 2 bytes for faster SD perf */
				printf(" -> fail\n");
				return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_CARD);
			}
			
			if (!fragAck(NULL, rxSlkHdr, fi, mTxRxBuf + 1, SD_BLOCK_SIZE + 1))
				printf("cannot send rdata-ack\n");
		
			break;
		
		case NOTICE_XFER:
			if (len != SD_BLOCK_SIZE + 1)
				return;
			
			//verify we are in write mode
			if (mState.curOp != SdWriteInProgress)
				return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_STATE);
			
			//if not dupe, account for and write it, else just ack it
			if (!isDupe) {
				
				if (mState.doneSec >= mState.numSec)
					return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_SEQ2);
				
				secNow = mState.firstSec + mState.doneSec++;
				
	//			printf("**WR %u\n", secNow);
				if (!sdSecWrite(secNow, mTxRxBuf + 1)) {
					printf(" -> fail\n");
					return sasdPrvSendErrorAck(rxSlkHdr, fi, SD_ERR_CARD);
				}
			}
			else
				printf("ignoring dupe write\n");
			
			if (!fragAck(NULL, rxSlkHdr, fi, &ok, sizeof(ok)))
				printf("cannot send wdata-ack\n");
			break;
		
		case CMD_RESET_STATE:
			mState.curOp = SdNoneInProgress;
			if (!fragAck(NULL, rxSlkHdr, fi, &ok, sizeof(ok)))
				printf("cannot send wdata-ack\n");
			break;
		
		case CMD_SET_BAUD:
			if (len != sizeof(struct SdBaudData))
				return;
			
			printf("offered %u bps\n", TO_U32(baud->rate));
			
			if (TO_U32(baud->rate) > OUR_BAUDRATE_MAX)
				baud->rate = TO_U32(OUR_BAUDRATE_MAX);
			
			printf("will use %lu bps\n", TO_U32(baud->rate));
			
			baud->cmd = RSP_OK;
			if (!fragAck(NULL, rxSlkHdr, fi, baud, sizeof(struct SdBaudData)))
				printf("cannot send baud-ack\n");
		
			//wait for tx
			uartTxWait();
			
			//now change rate
			uartSetBaudrate(TO_U32(baud->rate));
			break;
		
		case CMD_GET_BATT:
			if (len != 1)
				return;
			
			battVolts = getBattVoltage();
			batt->cmd = RSP_OK;
			batt->percent = battVoltsToPercent(battVolts);
			batt->voltage = TO_U16(battVolts);
			
			if (!fragAck(NULL, rxSlkHdr, fi, batt, sizeof(struct SdBatteryData)))
				printf("cannot send batt-ack\n");
			break;
		
		case CMD_FW_UPDATE_START:
		
			if (!fragAck(NULL, rxSlkHdr, fi,  &ok, sizeof(ok)))
				printf("cannot send ufwupd-ack\n");
			
			selfErase();
			break;
		
		default:
			printf("unknown command 0x%02x\n", mTxRxBuf[0]);
			return;
	}
}





















static bool hotsyncPrvTxFrag(void *src, uint_fast16_t len)
{
	struct SlkHdr hdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, };
	
	return fragSend(NULL, &hdr, src, len, NULL);
}


static void hotsyncEnd(uint16_t endCode)
{
	if (hotsyncEndSync(endCode)) {
		
		//printf("sync ended\n");
	}
}

bool hsPrvFatReadSec(void *userData, uint32_t sec, void *dst)
{
	return sdSecRead(sec, dst);
}

static bool hsPrvPreinst(void)
{
	static const char *mPath[] = {"PALM", "PROGRAMS", "SASD", };
	bool success = true, ret = false;
	struct FatFileOpenInfoOpaque fio;
	struct FatFile *dir, *f;
	struct FatVolume *v;
	uint_fast16_t i;
	uint32_t size;
	char name[13];
	uint8_t attr;
	
	
	
	fatInit(mTxRxBuf);
	
	
	v = fatMount(hsPrvFatReadSec, NULL);
	if (!v) {
		printf("mount fail\n");
		goto out_unmount;
	}
	
	dir = fatOpenRoot(v);
	if (!dir) {
		printf("root dir open fail\n");
		goto out_unmount;
	}
	
	for (i = 0; i < sizeof(mPath) / sizeof(*mPath) && success; i++) {
		
		success = false;
		while (fatDirEnum(dir, name, &attr, NULL, &fio)) {
			
			if ((attr & FAT_FLAG_DIR) && !strcmp(mPath[i], name)) {
				
				fatFileClose(dir);
				dir = fatFileOpen(v, &fio);
				success = !!dir;
				break;
			}
		}
	}
	
	if (success) {
		
		ret = true;
		while (fatDirEnum(dir, name, &attr, &size, &fio)) {
			
			printf("found file '%s'\n", name);
			
			if (!(attr & FAT_FLAG_DIR) && strlen(name) >= 4 && !strcmp(name + strlen(name) - 4, ".PRC")) {
				
				struct HsDbCreateReq dbcReq;
				uint16_t attr, nRes;
				uint32_t type, crid;
				int db;
				
				printf("file '%s'\n", name);
				
				f = fatFileOpen(v, &fio);
				if (!f) {
					
					printf("failed to open '%s'\n", name);
					ret = false;
					goto out_dirclose;
				}
				
				if (!prcOpen((PrcFileReadF)fatFileRead, f, size, dbcReq.dbName, &attr, &dbcReq.dbVersion, &type, &crid, &nRes))
					printf("failed to prc open\n");
				else {
					
					dbcReq.dbFlags = TO_U16(attr | PALM_DB_ATTR_RESET_AFTER_INSTALL);
					dbcReq.crid = TO_U32(crid);
					dbcReq.type = TO_U32(type);
					dbcReq.dbVersion = TO_U16(dbcReq.dbVersion);
					dbcReq.cardNo = 0;
					dbcReq.rfu = 0;
					
					int16_t db = hotsyncDbCreate(&dbcReq);
					
					if (db < 0)
						printf("db create fail\n");
					else {

						printf("created db '%s'\n", dbcReq.dbName);
												
						for (i = 0; i < nRes; i++) {
							
							uint32_t sendDone, sendNow;
							
							if (!prcBeginResRead(&type, &attr, &size)) {
								printf("db res open err\n");
								ret = false;
								goto out_prc_close;
							}
							
							printf("created res (%c%c%c%c, 0x%04x) sized %u\n",
								(type >> 24) & 0xff, (type >> 16) & 0xff, (type >> 8) & 0xff, (type >> 0) & 0xff, attr, size);
							
							if (!hotsyncResWriteBegin(db, type, attr, size)) {
								printf("res write start failed\n");
								ret = false;
								goto out_prc_close;
							}
							
							for (sendDone = 0; sendDone < size; sendDone += sendNow) {
								
								sendNow = size - sendDone;
								if (sendNow > sizeof(mHsDataBuffer))
									sendNow = sizeof(mHsDataBuffer);
								
								sendNow = prcContinueResRead(mHsDataBuffer, sendNow);
								
								printf("pos %u/%u, send %u\n",
									sendDone, size, sendNow);
								
								if (!hotsyncResWriteContinue(mHsDataBuffer, sendNow)) {
									printf("res write chunk failed\n");
									ret = false;
									goto out_prc_close;
								}
								printf("-> send done\n");
							}
							
							printf("->res write done\n");
							if (!hotsyncResWriteFinish()) {
								
								printf("res write end failed\n");
								ret = false;
								goto out_prc_close;
							}
						}
						printf("->db write done\n");
					}
					
					if (!hotsyncDbClose(db))
						printf("failed to close db");
					printf("->db closed\n");
					
		out_prc_close:
					prcClose();
				}
				fatFileClose(f);
			}
		}
		
		printf("enum done\n");
	}
	
out_dirclose:
	if (dir)
		fatFileClose(dir);

out_unmount:
	fatUnmount(v);
	return ret;
}

static void tryHotsync(void *pkt, uint16_t len)
{
	uint32_t osVer = 0, br = hotsyncInit(pkt, len, 115200);
	enum HsSyncEndCause endCause = HsSyncEndCauseOther;
	uint_fast8_t i, dbH;
	
	//we get here when we got first hotsync packet
	
	if (!br) {
		//printf("hotsync init failed\n");
		return;
	}
	
	uartSetBaudrate(br);
	
	printf("hotsync up at %u baud\n", (unsigned)br);
	
	//now we become master in the connection
	
	//update the ui to "syncing"
	if (!hotsyncOpenConduit())
		printf("failed 'open conduit' step\n");
	
	//get info (hack it to use less space)
	if (sizeof(osVer) != hotsyncGetSysInfo((struct HsSysInfo*)&osVer, sizeof(osVer)))
		osVer = 0;
	osVer = TO_U32(osVer);
	printf("Remote OS ver %u.%u.%u.%u.%u\n", osVer >> 24, (osVer >> 20) & 15, (osVer >> 16) & 15, (osVer >> 12) & 15, osVer & 0x0fff);
		
	printf("conduit open - start sync\n");
	
	if (hsPrvPreinst()) {
		
		if (osVer >= 0x02000000)
			endCause = HsSyncEndCauseNormal;
		else {
			
			printf("OS 1.0 will not reset by itself - make it!\n");
			if (hotsyncDemandResetAtEnd())
				endCause = HsSyncEndCauseNormal;
		}
	}
	
	hotsyncEnd(endCause);
}


///we care about code size, so we implement some smaller funcs, even if they are slow

//use rom divide api
uint64_t __attribute__((used)) __aeabi_uidivmod(uint32_t num, uint32_t denom)
{
	return ROM->div->uidiv_full(num, denom);
}
uint32_t __attribute__((weak, alias("__aeabi_uidivmod"), used)) __aeabi_uidiv(uint32_t num, uint32_t denom);

size_t strlen(const char *s)
{
	const char *t = s;
	
	while (*t)
		t++;
	
	return t - s;
}

int strcmp(const char *s1, const char *s2)
{
	char c1, c2;
	
	do {
		c1 = *s1++;
		c2 = *s2++;
	} while (c1 && c2 && c1 == c2);
	
	return c1 - c2;
}

void* memcpy(void *dstP, const void *srcP, size_t len)
{
	const uint8_t *src = (const uint8_t*)srcP;
	uint8_t *dst = (uint8_t*)dstP;
	
	while (len--)
		*dst++ = *src++;
	
	return dstP;
}

void* memset(void *dstP, int val, size_t len)
{
	uint8_t *dst = (uint8_t*)dstP;
	
	while (len--)
		*dst++ = val;
	
	return dstP;
}

char* strcpy(char *dst, const char *src)
{
	char *dstOrig = dst;
	
	while(*dst++ = *src++);
	
	return dstOrig;
}



